import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.io.*;
import javax.imageio.*;
import javax.swing.JOptionPane;
import javax.swing.JFileChooser;

/*
 * Created on 01.05.2004
 *
 */


/** 
 * 
 * @author Ruediger Knoerig
 */
public class FractalViewWidget extends Canvas implements Runnable {
	/**
	 *  
	 */
	public FractalViewWidget(JFractalView parent) 
	{
		this.parent=parent;
		setCursor(idlecursor = new Cursor(Cursor.CROSSHAIR_CURSOR));
		waitcursor = new Cursor(Cursor.WAIT_CURSOR);
		setFocusable(true);
		renderthread = new Thread(this);
		renderingcontrol = new Semaphore(1);
		enableEvents(AWTEvent.COMPONENT_EVENT_MASK | AWTEvent.MOUSE_EVENT_MASK | AWTEvent.MOUSE_MOTION_EVENT_MASK);
		setSize(400,400);
		renderthread.start();
		resetView();
		setIterationDepth(200);
	}
	
	/**
	 * destructor, stops thread execution.
	 */
	public void finalize()
	{
		shouldRender=shouldRun=false;
	}
	
	public Dimension getPreferredSize() { return new Dimension(400,400); }
	
	/**
	 * reset the fractal to display the whole mandelbrot set and
	 * starts recalculation of the image 
	 */
	final public void resetView() 
	{ 
		ul_real=-2.0; 
		lr_real=0.5; ul_imag=1.5; 
		lr_imag=-1.5; 
		recalculateStepSizes(); 
		recalculateImage(); 
		//modified=false;
	}
	
	/**
	 * Setter for the iteration depth. 
	 * Triggers recalculation of the fractal.
	 * @param iteration_depth new iteration depth
	 */
	final public void setIterationDepth(int iteration_depth)
	{
		it=iteration_depth;
		factor = 0xFFFFFFFFl/it;
		recalculateImage();
	}
	
	/**
	 * Getter for the current iteration depth
	 * @return current iteration depth
	 */
	final public int getIterationDepth() { return it; }
	
	final public double getUpperLeftReal() { return ul_real; }
	
	final public double getUpperLeftImag() { return ul_imag; }
	
	final public double getLowerRightReal() { return lr_real; }
	
	final public double getLowerRightImag() { return lr_imag; }
	
	/**
	 * take-a-screenshot action with GUI interaction. 
	 */
	final public void takeScreenShot()
	{
		synchronized(fractal)
		{
			JFileChooser dlg = new JFileChooser();
			dlg.setFileFilter(new javax.swing.filechooser.FileFilter(){
				public boolean accept(File dir) { return dir.canWrite() && dir.getName().endsWith(".jpg"); }
				public String getDescription() { return "JPEG files"; }
			});
			dlg.setFileFilter(new javax.swing.filechooser.FileFilter(){
				public boolean accept(File dir) { return dir.canWrite() && dir.getName().endsWith(".png"); }
				public String getDescription() { return "PNG files"; }
			});
			
			if(dlg.showSaveDialog(parent) == JFileChooser.APPROVE_OPTION)
			{	
				try{
					String filtername = dlg.getFileFilter().getDescription();
					if(filtername == "PNG files") ImageIO.write(fractal,"png", dlg.getSelectedFile()); 
					else if(filtername == "JPEG files") ImageIO.write(fractal,"jpg", dlg.getSelectedFile());
					else JOptionPane.showMessageDialog(this,"Unknown image format!","JFractalView",JOptionPane.ERROR_MESSAGE);					
				}
				catch(IOException e)
				{
					JOptionPane.showMessageDialog(this,"Could not write screenshot image:\n"+e.getMessage(),"JFractalView",JOptionPane.ERROR_MESSAGE); 	
				}
			}
		}
	}
	
	
    /**
     * zoom in action
     */
    final public void zoomIn()
    {
     shouldRender=false;
     synchronized(fractal)
	 {
     	double real = lr_real-ul_real;
     	lr_real = ul_real + 0.9*real;
     	ul_real = ul_real + 0.1*real;
     	double imag = ul_imag-lr_imag;
     	lr_imag = ul_imag - 0.9*imag;
     	ul_imag = ul_imag - 0.1*imag;    
     	recalculateStepSizes();
	 }
     recalculateImage();
    }
    
    /**
     * zoom out action
     */
    void zoomOut()
    {
    	shouldRender=false;
    	synchronized(fractal)
		{
    		double real = 0.1*(lr_real-ul_real);
    		lr_real = lr_real + real;
    		ul_real = ul_real - real;
    		double imag = 0.1*(ul_imag-lr_imag);
    		lr_imag = lr_imag - imag;
    		ul_imag = ul_imag + imag;    
    		recalculateStepSizes();
		}
    	recalculateImage();
    }
	
	final public void setViewedArea(double ul_real, double ul_imag, double lr_real, double lr_imag, int iteration_depth)
	{
		shouldRender=false;
		synchronized(fractal)
		{
			if(ul_real > lr_real)
			{
				this.ul_real = lr_real;
				this.lr_real = ul_real;
			}
			else
			{
				this.ul_real = ul_real;
				this.lr_real = lr_real;
			}		
			
			if(ul_imag > lr_imag)
			{
				this.ul_imag = ul_imag;
				this.lr_imag = lr_imag;
			}
			else
			{
				this.ul_imag = lr_imag;
				this.lr_imag = ul_imag;
			}		
			recalculateStepSizes();
		}
		setIterationDepth(iteration_depth);
	}
	
	/**
	 * Thread run function, performs rendering
	 */
	public void run()
	{
		while(shouldRun)
		{
			try{
				//System.err.println("New renderthread loop run.");
				renderingcontrol.acquire();
				if(!shouldRun) return;
				if(fractal == null) continue;
				//System.err.println("Rendering an "+getWidth()+"x"+getHeight()+" fractal!");
				//System.err.println("Image is "+fractal.getWidth()+"x"+fractal.getHeight());
				long time;
				synchronized(fractal)
				{
					setCursor(waitcursor);
					time=System.currentTimeMillis();
					DataBufferInt dst = (DataBufferInt)(fractal.getRaster().getDataBuffer());
					double ci = ul_imag;
					int line_inc = fractal.getWidth();
					int dptr = 0;
					int line_end = 0;
					int eptr = line_inc*fractal.getHeight();
					while(shouldRender && dptr != eptr)
					{
						double cr = ul_real;
						line_end += line_inc;
						while(dptr != line_end)
						{
							double zr = cr; 
							double zqr = cr*cr;
							double zi = ci; double zqi = ci*ci;
							int i = it;
							do
							{
								zi = 2*zr*zi + ci;
								zr = zqr-zqi + cr;
								zqr = zr*zr;
								zqi = zi*zi;
							}while(--i != 0 && (zqr+zqi < 4.0));
							dst.setElem(dptr++,(int) (i*factor | 0xFF000000));
							cr += dr;
						}
						ci -= di; 
					} 
				}
				if(shouldRender)
				{
					time=System.currentTimeMillis()-time;
					setParentStatusText("Rendered a "+fractal.getWidth()+"x"+fractal.getHeight()+" (IT="+it+") fractal in "+time+" msec!");
					paint(getGraphics());
				}
				setCursor(idlecursor);
			}
			catch(InterruptedException e)
			{
				return;
			}
		}
	}
	
	/** 
	 * reimplementation of Canvas.update(Graphics g) preventing
	 */
	public void update(Graphics g)
	{
		paint(g);
	}
	
	/**
	 * paints the fractal to the canvas
	 */
	public void paint(Graphics g)
	{
		if(fractal == null || g == null) return;
		synchronized(fractal)
		{
			g.drawImage(fractal,0,0,this);
			//System.err.println("Draw a "+fractal.getWidth()+"x"+fractal.getHeight()+" Image");
		}
	}
	
	final private void recalculateImage() 
	{ 
		if(fractal != null) 
		{ 
			shouldRender=false;
			synchronized(fractal)
			{
				shouldRender=true;
				renderingcontrol.release();
			}
		} 
	}
	
	final private void setParentStatusText(String text) { parent.statusBar.setText(text); }
	
	protected void processComponentEvent(ComponentEvent e)
	{
		
		if(e.getID() == ComponentEvent.COMPONENT_RESIZED)
		{
			
			if(fractal != null)
			{
				shouldRender = false;
				synchronized(fractal)
				{
					fractal = new BufferedImage(getWidth(),getHeight(),BufferedImage.TYPE_INT_RGB);
				}
			}
			else fractal = new BufferedImage(getWidth(),getHeight(),BufferedImage.TYPE_INT_RGB);
			recalculateStepSizes();
			recalculateImage();	
		}
	}
	
	protected void processMouseEvent(MouseEvent e)
	{
		switch(e.getID())
		{
			case MouseEvent.MOUSE_PRESSED:
				mousedown = true;
				firstclick = lastclick = e.getPoint();
				break;
			case MouseEvent.MOUSE_RELEASED:
				mousedown=false;
				lastclick = e.getPoint();
				double lrreal = ul_real+lastclick.x*dr;
				double lrimag = ul_imag-lastclick.y*di;
				double ulreal = ul_real+firstclick.x*dr;
				double ulimag = ul_imag-firstclick.y*di;	
				setViewedArea(ulreal,ulimag,lrreal,lrimag,getIterationDepth());
				break;
		}
	}
	
	protected void processMouseMotionEvent(MouseEvent e)
	{
		switch(e.getID())
		{
			case MouseEvent.MOUSE_DRAGGED:
				Graphics g=getGraphics();
				g.setXORMode(Color.white);
				g.drawRect(firstclick.x,firstclick.y,lastclick.x-firstclick.x,lastclick.y-firstclick.y);
				lastclick=e.getPoint();
				g.drawRect(firstclick.x,firstclick.y,lastclick.x-firstclick.x,lastclick.y-firstclick.y);
			case MouseEvent.MOUSE_MOVED:
				Point p=e.getPoint();
				setParentStatusText((ul_real+p.x*dr)+" + i * "+(ul_imag+p.y*di));
				break;
		}
	}	
	
	/**
	 * Recalculates the step sizes from one point in the complex plane to the next point.
	 */ 	
	final private void recalculateStepSizes() 
	{ 
		if(fractal != null) 
		{ 
			dr = java.lang.Math.abs(lr_real-ul_real)/fractal.getWidth(); 
			di = java.lang.Math.abs(ul_imag-lr_imag)/fractal.getHeight();   
		} 
	} 
	
	/** thread for running the rendering job */
	private Thread renderthread; 
	/** fractal image */
	private BufferedImage fractal;
	/** real part of the fractal's upper-left edge in the complex plane */ 
	private double ul_real; 
	/** imaginary part of the fractal's upper-left edge in the complex plane */ 
	private double ul_imag; 
	/** real part of the fractal's lower-right edge in the complex plane */ 
	private double lr_real; 
	/** imaginary part of the fractal's upper-left edge in the complex plane */ 
	private double lr_imag;
	/** step size to the next fractal point along the real axis */
	double dr;
	/** step size to the next fractal point along the imaginary axis */
	double di;
	/** iteration depth */
	int  it = 200;
	/** shift factor for color calculation */
	long factor = 10;
	/** flag to indicate if the thread should run */
	boolean shouldRun=true;
	/** flag indicating if the thread should process on rendering the image */
	boolean shouldRender=true;
	/** indicating new rendering jobs */
	Semaphore renderingcontrol;
	/** position of first mouse click */
	Point firstclick;
	/** position of last mouse click */
	Point lastclick;
	/** indicating if the mouse is down */
	boolean mousedown = false;
	/** reference to the parent widget */
	JFractalView parent;
	/** wait cursor */
	Cursor waitcursor;
	/** idle cursor */
	Cursor idlecursor;
}
